#import "ABCController.h"
#import <AudioToolbox/AudioToolbox.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <sys/time.h>

#define LS(str) NSLocalizedString(str,@"")
#ifdef _BIG_ENDIAN
#define SWAP32(n) (n)
#define SWAP16(n) (n)
#else
#define SWAP32(n) (((n>>24)&0xff) | ((n>>8)&0xff00) | ((n<<8)&0xff0000) | ((n<<24)&0xff000000))
#define SWAP16(n) (((n>>8)&0xff) | ((n<<8)&0xff00))
#endif

@implementation ABCController

- (id)init
{
	[super init];
	fileArray = [[NSMutableArray alloc] init];
	errorArray = [[NSMutableArray alloc] init];
	launched = NO;
	running = NO;
	firstDrag = YES;
	cancel = NO;
	return self;
}

- (IBAction)begin:(id)sender
{
	running = YES;
	[errorArray removeAllObjects];
	[o_mainWindow close];
	[o_progressWindow center];
	[o_progressWindow makeKeyAndOrderFront:nil];
	[o_progressBar setMaxValue:[fileArray count]];
	[o_progressBar setMinValue:0];
	[o_progressBar setDoubleValue:0];
	[NSThread detachNewThreadSelector:@selector(changeBitrate) toTarget:self withObject:nil];
}

- (IBAction)cancel:(id)sender
{
	cancel = YES;
}

- (void)changeBitrate
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	int i,j;
	for(i=0;i<[fileArray count] && !cancel;[o_progressBar setDoubleValue:++i]) {
		ExtAudioFileRef file;
		FSRef inputFSRef;
		FILE *fp;
		int tmp;
		char atom[4];
		struct stat fileStat;
		struct timeval tv[2];
		
		FSPathMakeRef((UInt8 *)[[fileArray objectAtIndex:i] UTF8String],&inputFSRef,NULL);
		if(ExtAudioFileOpen(&inputFSRef, &file) != noErr) {
			[errorArray addObject:[[fileArray objectAtIndex:i] stringByAbbreviatingWithTildeInPath]];
			continue;
		}
		AudioStreamBasicDescription inputFormat;
		UInt32 size = sizeof(inputFormat);
		if(ExtAudioFileGetProperty(file, kExtAudioFileProperty_FileDataFormat, &size, &inputFormat) != noErr) {
			ExtAudioFileDispose(file);
			[errorArray addObject:[[fileArray objectAtIndex:i] stringByAbbreviatingWithTildeInPath]];
			continue;
		}
		SInt64 frames;
		size = sizeof(frames);
		if(ExtAudioFileGetProperty(file, kExtAudioFileProperty_FileLengthFrames, &size, &frames) != noErr) {
			ExtAudioFileDispose(file);
			[errorArray addObject:[[fileArray objectAtIndex:i] stringByAbbreviatingWithTildeInPath]];
			continue;
		}
		if(inputFormat.mFormatID != kAudioFormatMPEG4AAC) {
			ExtAudioFileDispose(file);
			[errorArray addObject:[[fileArray objectAtIndex:i] stringByAbbreviatingWithTildeInPath]];
			continue;
		}
		double sec = (double)frames / (double)inputFormat.mSampleRate;
		ExtAudioFileDispose(file);
		
		stat([[fileArray objectAtIndex:i] UTF8String],&fileStat);
		fp = fopen([[fileArray objectAtIndex:i] UTF8String], "r+b");
		
		while(1) { //skip until mdat;
			if(fread(&tmp,4,1,fp) < 1) goto end;
			if(fread(atom,1,4,fp) < 4) goto end;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"mdat",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
		}
		unsigned int mdatSize = (unsigned int)tmp-8;
		int bitrate = round((tmp-8)/sec*8) ;
		//NSLog(@"%@: %f,%d,%lld",[fileArray objectAtIndex:i],sec,bitrate,frames);
		//NSLog(@"frame:%lld",frames);
		rewind(fp);
		
		while(1) { //skip until moov;
			if(fread(&tmp,4,1,fp) < 1) goto end;
			if(fread(atom,1,4,fp) < 4) goto end;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"moov",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
		}
		
		while(1) { //skip until trak;
			if(fread(&tmp,4,1,fp) < 1) goto end;
			if(fread(atom,1,4,fp) < 4) goto end;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"trak",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
		}
		
		long pos_trak = ftell(fp);
		int trakSize = tmp;
		
		while(1) { //skip until mdia;
			if(fread(&tmp,4,1,fp) < 1) goto end;
			if(fread(atom,1,4,fp) < 4) goto end;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"mdia",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
		}
		
		while(1) { //skip until minf;
			if(fread(&tmp,4,1,fp) < 1) goto end;
			if(fread(atom,1,4,fp) < 4) goto end;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"minf",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
		}
		
		while(1) { //skip until stbl;
			if(fread(&tmp,4,1,fp) < 1) goto end;
			if(fread(atom,1,4,fp) < 4) goto end;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"stbl",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
		}
		
		long pos_stbl = ftell(fp);
		
		unsigned int framesFromFile = 0;
		while(1) { //skip until stsz;
			if(fread(&tmp,4,1,fp) < 1) goto calc_bitrate;
			if(fread(atom,1,4,fp) < 4) goto calc_bitrate;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"stsz",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto calc_bitrate;
		}
		
		if(fseek(fp,8,SEEK_CUR) != 0) goto end;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		framesFromFile = (unsigned int)(SWAP32(tmp))*1024;
		
calc_bitrate:
		if(fseek(fp,pos_stbl,SEEK_SET) != 0) goto end;	
		if(framesFromFile) {
			sec = (double)framesFromFile / (double)inputFormat.mSampleRate;
			bitrate = round(mdatSize/sec*8) ;
		}
		
		while(1) { //skip until esds;
			if(fread(atom,1,4,fp) < 4) goto end;
			if(!memcmp(atom,"esds",4)) break;
			if(fseek(fp,-3,SEEK_CUR) != 0) goto end;
		}
		
		if(fseek(fp,5,SEEK_CUR) != 0) goto end;
		for(j=0;j<3;j++) {
			if(fread(atom,1,1,fp) < 1) goto end;
			if((unsigned char)atom[0] != 0x80) break;
		}
		if(fseek(fp,5,SEEK_CUR) != 0) goto end;
		for(j=0;j<3;j++) {
			if(fread(atom,1,1,fp) < 1) goto end;
			if((unsigned char)atom[0] != 0x80) break;
		}
		if(fseek(fp,10,SEEK_CUR) != 0) goto end;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		tmp = SWAP32(tmp);
		//NSLog(@"written bitrate is:%d",tmp);
		if(fseek(fp,-4,SEEK_CUR) != 0) goto end;
		
		if([[o_processMode selectedCell] tag] == 0) {
			bitrate = SWAP32(bitrate);
			if(fwrite(&bitrate,4,1,fp) < 1) goto end;
		}
		else if([[o_processMode selectedCell] tag] == 1) {
			bitrate = [o_bitrate intValue]*1000;
			bitrate = SWAP32(bitrate);
			if(fwrite(&bitrate,4,1,fp) < 1) goto end;
		}
		else {
			id bitrateArr = [o_bitrate objectValues];
			int diff,diff_min;
			diff_min = INT_MAX;
			
			for(j=0;j<[bitrateArr count];j++) {
				diff = abs(bitrate - [[bitrateArr objectAtIndex:j] intValue]*1000);
				if(diff < diff_min) diff_min = diff;
				else break;
			}
			//NSLog(@"actual: %d",bitrate);
			bitrate = [[bitrateArr objectAtIndex:j-1] intValue]*1000;
			//NSLog(@"selected: %d",bitrate);
			bitrate = SWAP32(bitrate);
			if(fwrite(&bitrate,4,1,fp) < 1) goto end;
		}
		
		if(fseek(fp,pos_trak,SEEK_SET) != 0) goto end;
		if(fseek(fp,trakSize-8,SEEK_CUR) != 0) goto end;
		
		while(1) { //skip until udta;
			if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
			if(fread(atom,1,4,fp) < 4) goto end_noerror;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"udta",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end_noerror;
		}
		
		while(1) { //skip until meta;
			if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
			if(fread(atom,1,4,fp) < 4) goto end_noerror;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"meta",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end_noerror;
		}
		
		if(fseek(fp,4,SEEK_CUR) != 0) goto end_noerror;
		
		while(1) { //skip until ilst;
			if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
			if(fread(atom,1,4,fp) < 4) goto end_noerror;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"ilst",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end_noerror;
		}
		
		while(1) {
			while(1) { //skip until ----;
				if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
				if(fread(atom,1,4,fp) < 4) goto end_noerror;
				tmp = SWAP32(tmp);
				if(!memcmp(atom,"----",4)) break;
				if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end_noerror;
			}
			
			long nextAtom = ftell(fp) - 8 + tmp;
			
			while(1) { //skip until name;
				if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
				if(fread(atom,1,4,fp) < 4) goto end_noerror;
				tmp = SWAP32(tmp);
				if(!memcmp(atom,"name",4)) break;
				if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end_noerror;
			}
			if(fseek(fp,4,SEEK_CUR) != 0) goto end_noerror;
			char *nameStr = (char *)malloc((tmp-12) < 15 ? 15 : tmp-12);
			if(fread(nameStr,1,tmp-12,fp) < tmp-12) goto end_noerror;
			if(!memcmp(nameStr,"Encoding Params",15)) {
				free(nameStr); 
				break;
			}
			free(nameStr); 
			if(fseek(fp,nextAtom,SEEK_SET) != 0) goto end_noerror;
		}
		
		while(1) { //skip until data;
			if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
			if(fread(atom,1,4,fp) < 4) goto end_noerror;
			tmp = SWAP32(tmp);
			if(!memcmp(atom,"data",4)) break;
			if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end_noerror;
		}
		
		if(fseek(fp,8,SEEK_CUR) != 0) goto end_noerror;
		int dataRest = tmp - 16;
		while(dataRest) {
			if(fread(atom,1,4,fp) < 4) goto end_noerror;
			if(!memcmp(atom,"brat",4)) {
				if(fread(&tmp,4,1,fp) < 1) goto end_noerror;
				tmp = SWAP32(tmp);
				if(fseek(fp,-4,SEEK_CUR) != 0) goto end;
				if(fwrite(&bitrate,4,1,fp) < 1) goto end;
			}
			else if(fseek(fp,4,SEEK_CUR) != 0) goto end_noerror;
			dataRest -= 8;
		}
		
end_noerror:
		fclose(fp);
		
		//NSLog(@"%d",fileStat.st_mtimespec.tv_nsec);
		if([o_remainModifiedDate state] == NSOnState) {
			tv[0].tv_sec = fileStat.st_atimespec.tv_sec;
			tv[0].tv_usec = fileStat.st_atimespec.tv_nsec/1000;
			tv[1].tv_sec = fileStat.st_mtimespec.tv_sec;
			tv[1].tv_usec = fileStat.st_mtimespec.tv_nsec/1000;
			utimes([[fileArray objectAtIndex:i] UTF8String],tv);
		}
		
		continue;
		
end:
		[errorArray addObject:[[fileArray objectAtIndex:i] stringByAbbreviatingWithTildeInPath]];
		fclose(fp);
		if([o_remainModifiedDate state] == NSOnState) {
			tv[0].tv_sec = fileStat.st_atimespec.tv_sec;
			tv[0].tv_usec = fileStat.st_atimespec.tv_nsec/1000;
			tv[1].tv_sec = fileStat.st_mtimespec.tv_sec;
			tv[1].tv_usec = fileStat.st_mtimespec.tv_nsec/1000;
			utimes([[fileArray objectAtIndex:i] UTF8String],tv);
		}
		
	}
	running = NO;
	cancel = NO;
	[self performSelectorOnMainThread:@selector(showResult) withObject:nil waitUntilDone:NO];
	[pool release];
}

- (void)showResult
{
	[o_resultStr setStringValue:[NSString stringWithFormat:LS(@"Processed %d files in %d files"),[fileArray count]-[errorArray count],[fileArray count]]];
	[o_resultTable reloadData];
	[o_progressWindow close];
	[o_resultWindow center];
	[o_resultWindow makeKeyAndOrderFront:nil];
}

- (void)scanDirectory:(NSString *)dirPath
{
	int i;
	NSFileManager *mgr = [NSFileManager defaultManager];
	NSArray *arr = [mgr directoryContentsAtPath:dirPath];
	for(i=0;i<[arr count];i++) {
		NSString *file = [arr objectAtIndex:i];
		BOOL isDir;
		[mgr fileExistsAtPath:[dirPath stringByAppendingPathComponent:file] isDirectory:&isDir];
		if(isDir) {
			[self scanDirectory:[dirPath stringByAppendingPathComponent:file]];
			continue;
		}
		if([[[file pathExtension] lowercaseString] isEqualToString:@"m4a"] || [[[file pathExtension] lowercaseString] isEqualToString:@"mp4"])
			[fileArray addObject:[dirPath stringByAppendingPathComponent:file]];
	}
}

- (void)processQueue
{
	firstDrag = YES;
	if([fileArray count] == 1) {
		NSFileManager *mgr = [NSFileManager defaultManager];
		BOOL isDir;
		[mgr fileExistsAtPath:[fileArray objectAtIndex:0] isDirectory:&isDir];
		if(isDir) {
			[self scanDirectory:[fileArray objectAtIndex:0]];
			[fileArray removeObjectAtIndex:0];
		}
	}
	[o_mainWindow center];
	[o_mainWindow makeKeyAndOrderFront:nil];
}

- (int)numberOfRowsInTableView:(NSTableView *)tableView
{
	return [errorArray count];
}

- (id)tableView:(NSTableView *)tableView
		objectValueForTableColumn:(NSTableColumn *)tableColumn
			row:(int)row
{
	return [errorArray objectAtIndex:row];
}

- (BOOL)application:(NSApplication *)theApplication openFile:(NSString *)filename
{
	if(running) return NO;
	if(firstDrag && launched) {
		[o_resultWindow close];
		[fileArray removeAllObjects];
		firstDrag = NO;
		[self performSelector:@selector(processQueue) withObject:nil afterDelay:0.2];
	}
	[fileArray addObject:filename];
	return YES;
}

- (void)applicationDidFinishLaunching: (NSNotification *)notification
{
	[o_bitrate setIntValue:128];
	launched = YES;
	if([fileArray count]) [self processQueue];
}


@end
